@@ -1,6 +1,6 @@
 //  Copyright John Maddock 2007.
 //  Copyright Paul A. Bristow 2007.
-
+//  Copyright Matt Borland 2024.
 //  Use, modification and distribution are subject to the
 //  Boost Software License, Version 1.0. (See accompanying file
 //  LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
@@ -8,22 +8,32 @@
 #ifndef BOOST_MATH_POLICY_ERROR_HANDLING_HPP
 #define BOOST_MATH_POLICY_ERROR_HANDLING_HPP
 
-#include <boost/math/tools/config.hpp>
-#include <iomanip>
-#include <string>
-#include <cstring>
-#ifndef BOOST_MATH_NO_RTTI
-#include <typeinfo>
-#endif
-#include <cerrno>
-#include <complex>
-#include <cmath>
-#include <cstdint>
 #include <boost/math/policies/policy.hpp>
+#include <boost/math/tools/config.hpp>
+#include <boost/math/tools/numeric_limits.hpp>
 #include <boost/math/tools/precision.hpp>
+#include <boost/math/tools/tuple.hpp>
+#include <boost/math/tools/type_traits.hpp>
+
+#ifndef BOOST_MATH_HAS_NVRTC
+
+#ifndef BOOST_MATH_NO_EXCEPTIONS
+#include <boost/math/tools/throw_exception.hpp>
+#endif
+
+#include <cerrno>
+#include <cmath>
+#include <complex>
+#include <cstdint>
+#include <cstring>
+#include <iomanip>
+#include <sstream>
 #ifndef BOOST_MATH_NO_EXCEPTIONS
 #include <stdexcept>
-#include <boost/math/tools/throw_exception.hpp>
+#endif
+#include <string>
+#ifndef BOOST_MATH_NO_RTTI
+#include <typeinfo>
 #endif
 
 #ifdef _MSC_VER
@@ -36,7 +46,6 @@
 // Note that this only occurs when the compiler can deduce code is unreachable,
 // for example when policy macros are used to ignore errors rather than throw.
 #endif
-#include <sstream>
 
 namespace boost{ namespace math{
 
@@ -54,6 +63,11 @@ public:
    explicit rounding_error(const std::string& s) : std::runtime_error(s){}
 };
 
+#else
+
+class evaluation_error {};
+class rounding_error {};
+
 #endif
 
 namespace policies{
@@ -84,15 +98,20 @@ namespace detail
 template <class T>
 inline std::string prec_format(const T& val)
 {
-   typedef typename boost::math::policies::precision<T, boost::math::policies::policy<> >::type prec_type;
-   std::stringstream ss;
+   using prec_type = typename boost::math::policies::precision<T, boost::math::policies::policy<> >::type;
+
+   std::stringstream strm { };
+
    if(prec_type::value)
    {
-      int prec = 2 + (prec_type::value * 30103UL) / 100000UL;
-      ss << std::setprecision(prec);
+      const std::streamsize prec { static_cast<std::streamsize>(2UL + (prec_type::value * 30103UL) / 100000UL) };
+
+      strm << std::setprecision(prec);
    }
-   ss << val;
-   return ss.str();
+
+   strm << val;
+
+   return strm.str();
 }
 
 #ifdef BOOST_MATH_USE_CHARCONV_FOR_CONVERSION
@@ -199,7 +218,7 @@ void raise_error(const char* pfunction, const char* pmessage, const T& val)
 #endif
 
 template <class T>
-inline T raise_domain_error(
+BOOST_MATH_GPU_ENABLED inline T raise_domain_error(
            const char* function,
            const char* message,
            const T& val,
@@ -210,12 +229,12 @@ inline T raise_domain_error(
 #else
    raise_error<std::domain_error, T>(function, message, val);
    // we never get here:
-   return std::numeric_limits<T>::quiet_NaN();
+   return boost::math::numeric_limits<T>::quiet_NaN();
 #endif
 }
 
 template <class T>
-inline constexpr T raise_domain_error(
+BOOST_MATH_GPU_ENABLED constexpr T raise_domain_error(
            const char* ,
            const char* ,
            const T& ,
@@ -223,11 +242,11 @@ inline constexpr T raise_domain_error(
 {
    // This may or may not do the right thing, but the user asked for the error
    // to be ignored so here we go anyway:
-   return std::numeric_limits<T>::quiet_NaN();
+   return boost::math::numeric_limits<T>::quiet_NaN();
 }
 
 template <class T>
-inline T raise_domain_error(
+BOOST_MATH_GPU_ENABLED inline T raise_domain_error(
            const char* ,
            const char* ,
            const T& ,
@@ -236,11 +255,11 @@ inline T raise_domain_error(
    errno = EDOM;
    // This may or may not do the right thing, but the user asked for the error
    // to be silent so here we go anyway:
-   return std::numeric_limits<T>::quiet_NaN();
+   return boost::math::numeric_limits<T>::quiet_NaN();
 }
 
 template <class T>
-inline T raise_domain_error(
+BOOST_MATH_GPU_ENABLED inline T raise_domain_error(
            const char* function,
            const char* message,
            const T& val,
@@ -250,7 +269,7 @@ inline T raise_domain_error(
 }
 
 template <class T>
-inline T raise_pole_error(
+BOOST_MATH_GPU_ENABLED inline T raise_pole_error(
            const char* function,
            const char* message,
            const T& val,
@@ -264,7 +283,7 @@ inline T raise_pole_error(
 }
 
 template <class T>
-inline constexpr T raise_pole_error(
+BOOST_MATH_GPU_ENABLED constexpr T raise_pole_error(
            const char* function,
            const char* message,
            const T& val,
@@ -274,7 +293,7 @@ inline constexpr T raise_pole_error(
 }
 
 template <class T>
-inline constexpr T raise_pole_error(
+BOOST_MATH_GPU_ENABLED constexpr T raise_pole_error(
            const char* function,
            const char* message,
            const T& val,
@@ -284,7 +303,7 @@ inline constexpr T raise_pole_error(
 }
 
 template <class T>
-inline T raise_pole_error(
+BOOST_MATH_GPU_ENABLED inline T raise_pole_error(
            const char* function,
            const char* message,
            const T& val,
@@ -294,7 +313,7 @@ inline T raise_pole_error(
 }
 
 template <class T>
-inline T raise_overflow_error(
+BOOST_MATH_GPU_ENABLED inline T raise_overflow_error(
            const char* function,
            const char* message,
            const  ::boost::math::policies::overflow_error< ::boost::math::policies::throw_on_error>&)
@@ -304,12 +323,12 @@ inline T raise_overflow_error(
 #else
    raise_error<std::overflow_error, T>(function, message ? message : "numeric overflow");
    // We should never get here:
-   return std::numeric_limits<T>::has_infinity ? std::numeric_limits<T>::infinity() : boost::math::tools::max_value<T>();
+   return boost::math::numeric_limits<T>::has_infinity ? boost::math::numeric_limits<T>::infinity() : boost::math::tools::max_value<T>();
 #endif
 }
 
 template <class T>
-inline T raise_overflow_error(
+BOOST_MATH_GPU_ENABLED inline T raise_overflow_error(
            const char* function,
            const char* message,
            const T& val,
@@ -320,23 +339,23 @@ inline T raise_overflow_error(
 #else
    raise_error<std::overflow_error, T>(function, message ? message : "numeric overflow", val);
    // We should never get here:
-   return std::numeric_limits<T>::has_infinity ? std::numeric_limits<T>::infinity() : boost::math::tools::max_value<T>();
+   return boost::math::numeric_limits<T>::has_infinity ? boost::math::numeric_limits<T>::infinity() : boost::math::tools::max_value<T>();
 #endif
 }
 
 template <class T>
-inline constexpr T raise_overflow_error(
+BOOST_MATH_GPU_ENABLED constexpr T raise_overflow_error(
            const char* ,
            const char* ,
            const  ::boost::math::policies::overflow_error< ::boost::math::policies::ignore_error>&) BOOST_MATH_NOEXCEPT(T)
 {
    // This may or may not do the right thing, but the user asked for the error
    // to be ignored so here we go anyway:
-   return std::numeric_limits<T>::has_infinity ? std::numeric_limits<T>::infinity() : boost::math::tools::max_value<T>();
+   return boost::math::numeric_limits<T>::has_infinity ? boost::math::numeric_limits<T>::infinity() : boost::math::tools::max_value<T>();
 }
 
 template <class T>
-inline constexpr T raise_overflow_error(
+BOOST_MATH_GPU_ENABLED constexpr T raise_overflow_error(
            const char* ,
            const char* ,
            const T&,
@@ -344,11 +363,11 @@ inline constexpr T raise_overflow_error(
 {
    // This may or may not do the right thing, but the user asked for the error
    // to be ignored so here we go anyway:
-   return std::numeric_limits<T>::has_infinity ? std::numeric_limits<T>::infinity() : boost::math::tools::max_value<T>();
+   return boost::math::numeric_limits<T>::has_infinity ? boost::math::numeric_limits<T>::infinity() : boost::math::tools::max_value<T>();
 }
 
 template <class T>
-inline T raise_overflow_error(
+BOOST_MATH_GPU_ENABLED inline T raise_overflow_error(
            const char* ,
            const char* ,
            const  ::boost::math::policies::overflow_error< ::boost::math::policies::errno_on_error>&) BOOST_MATH_NOEXCEPT(T)
@@ -356,11 +375,11 @@ inline T raise_overflow_error(
    errno = ERANGE;
    // This may or may not do the right thing, but the user asked for the error
    // to be silent so here we go anyway:
-   return std::numeric_limits<T>::has_infinity ? std::numeric_limits<T>::infinity() : boost::math::tools::max_value<T>();
+   return boost::math::numeric_limits<T>::has_infinity ? boost::math::numeric_limits<T>::infinity() : boost::math::tools::max_value<T>();
 }
 
 template <class T>
-inline T raise_overflow_error(
+BOOST_MATH_GPU_ENABLED inline T raise_overflow_error(
            const char* ,
            const char* ,
            const T&,
@@ -369,20 +388,20 @@ inline T raise_overflow_error(
    errno = ERANGE;
    // This may or may not do the right thing, but the user asked for the error
    // to be silent so here we go anyway:
-   return std::numeric_limits<T>::has_infinity ? std::numeric_limits<T>::infinity() : boost::math::tools::max_value<T>();
+   return boost::math::numeric_limits<T>::has_infinity ? boost::math::numeric_limits<T>::infinity() : boost::math::tools::max_value<T>();
 }
 
 template <class T>
-inline T raise_overflow_error(
+BOOST_MATH_GPU_ENABLED inline T raise_overflow_error(
            const char* function,
            const char* message,
            const  ::boost::math::policies::overflow_error< ::boost::math::policies::user_error>&)
 {
-   return user_overflow_error(function, message, std::numeric_limits<T>::infinity());
+   return user_overflow_error(function, message, boost::math::numeric_limits<T>::infinity());
 }
 
 template <class T>
-inline T raise_overflow_error(
+BOOST_MATH_GPU_ENABLED inline T raise_overflow_error(
            const char* function,
            const char* message,
            const T& val,
@@ -392,11 +411,11 @@ inline T raise_overflow_error(
    std::string sval = prec_format(val);
    replace_all_in_string(m, "%1%", sval.c_str());
 
-   return user_overflow_error(function, m.c_str(), std::numeric_limits<T>::infinity());
+   return user_overflow_error(function, m.c_str(), boost::math::numeric_limits<T>::infinity());
 }
 
 template <class T>
-inline T raise_underflow_error(
+BOOST_MATH_GPU_ENABLED inline T raise_underflow_error(
            const char* function,
            const char* message,
            const  ::boost::math::policies::underflow_error< ::boost::math::policies::throw_on_error>&)
@@ -411,7 +430,7 @@ inline T raise_underflow_error(
 }
 
 template <class T>
-inline constexpr T raise_underflow_error(
+BOOST_MATH_GPU_ENABLED constexpr T raise_underflow_error(
            const char* ,
            const char* ,
            const  ::boost::math::policies::underflow_error< ::boost::math::policies::ignore_error>&) BOOST_MATH_NOEXCEPT(T)
@@ -422,7 +441,7 @@ inline constexpr T raise_underflow_error(
 }
 
 template <class T>
-inline T raise_underflow_error(
+BOOST_MATH_GPU_ENABLED inline T raise_underflow_error(
            const char* /* function */,
            const char* /* message */,
            const  ::boost::math::policies::underflow_error< ::boost::math::policies::errno_on_error>&) BOOST_MATH_NOEXCEPT(T)
@@ -434,7 +453,7 @@ inline T raise_underflow_error(
 }
 
 template <class T>
-inline T raise_underflow_error(
+BOOST_MATH_GPU_ENABLED inline T raise_underflow_error(
            const char* function,
            const char* message,
            const  ::boost::math::policies::underflow_error< ::boost::math::policies::user_error>&)
@@ -443,7 +462,7 @@ inline T raise_underflow_error(
 }
 
 template <class T>
-inline T raise_denorm_error(
+BOOST_MATH_GPU_ENABLED inline T raise_denorm_error(
            const char* function,
            const char* message,
            const T& /* val */,
@@ -459,7 +478,7 @@ inline T raise_denorm_error(
 }
 
 template <class T>
-inline constexpr T raise_denorm_error(
+BOOST_MATH_GPU_ENABLED inline constexpr T raise_denorm_error(
            const char* ,
            const char* ,
            const T&  val,
@@ -471,7 +490,7 @@ inline constexpr T raise_denorm_error(
 }
 
 template <class T>
-inline T raise_denorm_error(
+BOOST_MATH_GPU_ENABLED inline T raise_denorm_error(
            const char* ,
            const char* ,
            const T& val,
@@ -484,7 +503,7 @@ inline T raise_denorm_error(
 }
 
 template <class T>
-inline T raise_denorm_error(
+BOOST_MATH_GPU_ENABLED inline T raise_denorm_error(
            const char* function,
            const char* message,
            const T& val,
@@ -494,7 +513,7 @@ inline T raise_denorm_error(
 }
 
 template <class T>
-inline T raise_evaluation_error(
+BOOST_MATH_GPU_ENABLED inline T raise_evaluation_error(
            const char* function,
            const char* message,
            const T& val,
@@ -510,7 +529,7 @@ inline T raise_evaluation_error(
 }
 
 template <class T>
-inline constexpr T raise_evaluation_error(
+BOOST_MATH_GPU_ENABLED constexpr T raise_evaluation_error(
            const char* ,
            const char* ,
            const T& val,
@@ -522,7 +541,7 @@ inline constexpr T raise_evaluation_error(
 }
 
 template <class T>
-inline T raise_evaluation_error(
+BOOST_MATH_GPU_ENABLED inline T raise_evaluation_error(
            const char* ,
            const char* ,
            const T& val,
@@ -535,7 +554,7 @@ inline T raise_evaluation_error(
 }
 
 template <class T>
-inline T raise_evaluation_error(
+BOOST_MATH_GPU_ENABLED inline T raise_evaluation_error(
            const char* function,
            const char* message,
            const T& val,
@@ -545,7 +564,7 @@ inline T raise_evaluation_error(
 }
 
 template <class T, class TargetType>
-inline TargetType raise_rounding_error(
+BOOST_MATH_GPU_ENABLED inline TargetType raise_rounding_error(
            const char* function,
            const char* message,
            const T& val,
@@ -562,7 +581,7 @@ inline TargetType raise_rounding_error(
 }
 
 template <class T, class TargetType>
-inline constexpr TargetType raise_rounding_error(
+BOOST_MATH_GPU_ENABLED constexpr TargetType raise_rounding_error(
            const char* ,
            const char* ,
            const T& val,
@@ -571,12 +590,12 @@ inline constexpr TargetType raise_rounding_error(
 {
    // This may or may not do the right thing, but the user asked for the error
    // to be ignored so here we go anyway:
-   static_assert(std::numeric_limits<TargetType>::is_specialized, "The target type must have std::numeric_limits specialized.");
-   return  val > 0 ? (std::numeric_limits<TargetType>::max)() : (std::numeric_limits<TargetType>::is_integer ? (std::numeric_limits<TargetType>::min)() : -(std::numeric_limits<TargetType>::max)());
+   static_assert(boost::math::numeric_limits<TargetType>::is_specialized, "The target type must have std::numeric_limits specialized.");
+   return  val > 0 ? (boost::math::numeric_limits<TargetType>::max)() : (boost::math::numeric_limits<TargetType>::is_integer ? (boost::math::numeric_limits<TargetType>::min)() : -(boost::math::numeric_limits<TargetType>::max)());
 }
 
 template <class T, class TargetType>
-inline TargetType raise_rounding_error(
+BOOST_MATH_GPU_ENABLED inline TargetType raise_rounding_error(
            const char* ,
            const char* ,
            const T& val,
@@ -586,11 +605,11 @@ inline TargetType raise_rounding_error(
    errno = ERANGE;
    // This may or may not do the right thing, but the user asked for the error
    // to be silent so here we go anyway:
-   static_assert(std::numeric_limits<TargetType>::is_specialized, "The target type must have std::numeric_limits specialized.");
-   return  val > 0 ? (std::numeric_limits<TargetType>::max)() : (std::numeric_limits<TargetType>::is_integer ? (std::numeric_limits<TargetType>::min)() : -(std::numeric_limits<TargetType>::max)());
+   static_assert(boost::math::numeric_limits<TargetType>::is_specialized, "The target type must have std::numeric_limits specialized.");
+   return  val > 0 ? (boost::math::numeric_limits<TargetType>::max)() : (boost::math::numeric_limits<TargetType>::is_integer ? (boost::math::numeric_limits<TargetType>::min)() : -(boost::math::numeric_limits<TargetType>::max)());
 }
 template <class T, class TargetType>
-inline TargetType raise_rounding_error(
+BOOST_MATH_GPU_ENABLED inline TargetType raise_rounding_error(
            const char* function,
            const char* message,
            const T& val,
@@ -601,7 +620,7 @@ inline TargetType raise_rounding_error(
 }
 
 template <class T, class R>
-inline T raise_indeterminate_result_error(
+BOOST_MATH_GPU_ENABLED inline T raise_indeterminate_result_error(
            const char* function,
            const char* message,
            const T& val,
@@ -613,12 +632,12 @@ inline T raise_indeterminate_result_error(
 #else
    raise_error<std::domain_error, T>(function, message, val);
    // we never get here:
-   return std::numeric_limits<T>::quiet_NaN();
+   return boost::math::numeric_limits<T>::quiet_NaN();
 #endif
 }
 
 template <class T, class R>
-inline constexpr T raise_indeterminate_result_error(
+BOOST_MATH_GPU_ENABLED inline constexpr T raise_indeterminate_result_error(
            const char* ,
            const char* ,
            const T& ,
@@ -631,7 +650,7 @@ inline constexpr T raise_indeterminate_result_error(
 }
 
 template <class T, class R>
-inline T raise_indeterminate_result_error(
+BOOST_MATH_GPU_ENABLED inline T raise_indeterminate_result_error(
            const char* ,
            const char* ,
            const T& ,
@@ -645,7 +664,7 @@ inline T raise_indeterminate_result_error(
 }
 
 template <class T, class R>
-inline T raise_indeterminate_result_error(
+BOOST_MATH_GPU_ENABLED inline T raise_indeterminate_result_error(
            const char* function,
            const char* message,
            const T& val,
@@ -658,7 +677,7 @@ inline T raise_indeterminate_result_error(
 }  // namespace detail
 
 template <class T, class Policy>
-inline constexpr T raise_domain_error(const char* function, const char* message, const T& val, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
+BOOST_MATH_GPU_ENABLED constexpr T raise_domain_error(const char* function, const char* message, const T& val, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
 {
    typedef typename Policy::domain_error_type policy_type;
    return detail::raise_domain_error(
@@ -667,7 +686,7 @@ inline constexpr T raise_domain_error(const char* function, const char* message,
 }
 
 template <class T, class Policy>
-inline constexpr T raise_pole_error(const char* function, const char* message, const T& val, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
+BOOST_MATH_GPU_ENABLED constexpr T raise_pole_error(const char* function, const char* message, const T& val, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
 {
    typedef typename Policy::pole_error_type policy_type;
    return detail::raise_pole_error(
@@ -676,7 +695,7 @@ inline constexpr T raise_pole_error(const char* function, const char* message, c
 }
 
 template <class T, class Policy>
-inline constexpr T raise_overflow_error(const char* function, const char* message, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
+BOOST_MATH_GPU_ENABLED constexpr T raise_overflow_error(const char* function, const char* message, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
 {
    typedef typename Policy::overflow_error_type policy_type;
    return detail::raise_overflow_error<T>(
@@ -685,7 +704,7 @@ inline constexpr T raise_overflow_error(const char* function, const char* messag
 }
 
 template <class T, class Policy>
-inline constexpr T raise_overflow_error(const char* function, const char* message, const T& val, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
+BOOST_MATH_GPU_ENABLED constexpr T raise_overflow_error(const char* function, const char* message, const T& val, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
 {
    typedef typename Policy::overflow_error_type policy_type;
    return detail::raise_overflow_error(
@@ -694,7 +713,7 @@ inline constexpr T raise_overflow_error(const char* function, const char* messag
 }
 
 template <class T, class Policy>
-inline constexpr T raise_underflow_error(const char* function, const char* message, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
+BOOST_MATH_GPU_ENABLED constexpr T raise_underflow_error(const char* function, const char* message, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
 {
    typedef typename Policy::underflow_error_type policy_type;
    return detail::raise_underflow_error<T>(
@@ -703,7 +722,7 @@ inline constexpr T raise_underflow_error(const char* function, const char* messa
 }
 
 template <class T, class Policy>
-inline constexpr T raise_denorm_error(const char* function, const char* message, const T& val, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
+BOOST_MATH_GPU_ENABLED constexpr T raise_denorm_error(const char* function, const char* message, const T& val, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
 {
    typedef typename Policy::denorm_error_type policy_type;
    return detail::raise_denorm_error<T>(
@@ -713,7 +732,7 @@ inline constexpr T raise_denorm_error(const char* function, const char* message,
 }
 
 template <class T, class Policy>
-inline constexpr T raise_evaluation_error(const char* function, const char* message, const T& val, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
+BOOST_MATH_GPU_ENABLED constexpr T raise_evaluation_error(const char* function, const char* message, const T& val, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
 {
    typedef typename Policy::evaluation_error_type policy_type;
    return detail::raise_evaluation_error(
@@ -722,7 +741,7 @@ inline constexpr T raise_evaluation_error(const char* function, const char* mess
 }
 
 template <class T, class TargetType, class Policy>
-inline constexpr TargetType raise_rounding_error(const char* function, const char* message, const T& val, const TargetType& t, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
+BOOST_MATH_GPU_ENABLED constexpr TargetType raise_rounding_error(const char* function, const char* message, const T& val, const TargetType& t, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
 {
    typedef typename Policy::rounding_error_type policy_type;
    return detail::raise_rounding_error(
@@ -731,7 +750,7 @@ inline constexpr TargetType raise_rounding_error(const char* function, const cha
 }
 
 template <class T, class R, class Policy>
-inline constexpr T raise_indeterminate_result_error(const char* function, const char* message, const T& val, const R& result, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
+BOOST_MATH_GPU_ENABLED constexpr T raise_indeterminate_result_error(const char* function, const char* message, const T& val, const R& result, const Policy&) noexcept(is_noexcept_error_policy<Policy>::value && BOOST_MATH_IS_FLOAT(T))
 {
    typedef typename Policy::indeterminate_result_error_type policy_type;
    return detail::raise_indeterminate_result_error(
@@ -746,7 +765,7 @@ namespace detail
 {
 
 template <class R, class T, class Policy>
-BOOST_MATH_FORCEINLINE bool check_overflow(T val, R* result, const char* function, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T) && (Policy::value != throw_on_error) && (Policy::value != user_error))
+BOOST_MATH_GPU_ENABLED BOOST_MATH_FORCEINLINE bool check_overflow(T val, R* result, const char* function, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T) && (Policy::value != throw_on_error) && (Policy::value != user_error))
 {
    BOOST_MATH_STD_USING
    if(fabs(val) > tools::max_value<R>())
@@ -758,7 +777,7 @@ BOOST_MATH_FORCEINLINE bool check_overflow(T val, R* result, const char* functio
    return false;
 }
 template <class R, class T, class Policy>
-BOOST_MATH_FORCEINLINE bool check_overflow(std::complex<T> val, R* result, const char* function, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T) && (Policy::value != throw_on_error) && (Policy::value != user_error))
+BOOST_MATH_GPU_ENABLED BOOST_MATH_FORCEINLINE bool check_overflow(std::complex<T> val, R* result, const char* function, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T) && (Policy::value != throw_on_error) && (Policy::value != user_error))
 {
    typedef typename R::value_type r_type;
    r_type re, im;
@@ -768,7 +787,7 @@ BOOST_MATH_FORCEINLINE bool check_overflow(std::complex<T> val, R* result, const
    return r;
 }
 template <class R, class T, class Policy>
-BOOST_MATH_FORCEINLINE bool check_underflow(T val, R* result, const char* function, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T) && (Policy::value != throw_on_error) && (Policy::value != user_error))
+BOOST_MATH_GPU_ENABLED BOOST_MATH_FORCEINLINE bool check_underflow(T val, R* result, const char* function, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T) && (Policy::value != throw_on_error) && (Policy::value != user_error))
 {
    if((val != 0) && (static_cast<R>(val) == 0))
    {
@@ -778,7 +797,7 @@ BOOST_MATH_FORCEINLINE bool check_underflow(T val, R* result, const char* functi
    return false;
 }
 template <class R, class T, class Policy>
-BOOST_MATH_FORCEINLINE bool check_underflow(std::complex<T> val, R* result, const char* function, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T) && (Policy::value != throw_on_error) && (Policy::value != user_error))
+BOOST_MATH_GPU_ENABLED BOOST_MATH_FORCEINLINE bool check_underflow(std::complex<T> val, R* result, const char* function, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T) && (Policy::value != throw_on_error) && (Policy::value != user_error))
 {
    typedef typename R::value_type r_type;
    r_type re, im;
@@ -788,7 +807,7 @@ BOOST_MATH_FORCEINLINE bool check_underflow(std::complex<T> val, R* result, cons
    return r;
 }
 template <class R, class T, class Policy>
-BOOST_MATH_FORCEINLINE bool check_denorm(T val, R* result, const char* function, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T) && (Policy::value != throw_on_error) && (Policy::value != user_error))
+BOOST_MATH_GPU_ENABLED BOOST_MATH_FORCEINLINE bool check_denorm(T val, R* result, const char* function, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T) && (Policy::value != throw_on_error) && (Policy::value != user_error))
 {
    BOOST_MATH_STD_USING
    if((fabs(val) < static_cast<T>(tools::min_value<R>())) && (static_cast<R>(val) != 0))
@@ -799,7 +818,7 @@ BOOST_MATH_FORCEINLINE bool check_denorm(T val, R* result, const char* function,
    return false;
 }
 template <class R, class T, class Policy>
-BOOST_MATH_FORCEINLINE bool check_denorm(std::complex<T> val, R* result, const char* function, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T) && (Policy::value != throw_on_error) && (Policy::value != user_error))
+BOOST_MATH_GPU_ENABLED BOOST_MATH_FORCEINLINE bool check_denorm(std::complex<T> val, R* result, const char* function, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T) && (Policy::value != throw_on_error) && (Policy::value != user_error))
 {
    typedef typename R::value_type r_type;
    r_type re, im;
@@ -811,28 +830,28 @@ BOOST_MATH_FORCEINLINE bool check_denorm(std::complex<T> val, R* result, const c
 
 // Default instantiations with ignore_error policy.
 template <class R, class T>
-BOOST_MATH_FORCEINLINE constexpr bool check_overflow(T /* val */, R* /* result */, const char* /* function */, const overflow_error<ignore_error>&) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T))
+BOOST_MATH_GPU_ENABLED BOOST_MATH_FORCEINLINE constexpr bool check_overflow(T /* val */, R* /* result */, const char* /* function */, const overflow_error<ignore_error>&) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T))
 { return false; }
 template <class R, class T>
-BOOST_MATH_FORCEINLINE constexpr bool check_overflow(std::complex<T> /* val */, R* /* result */, const char* /* function */, const overflow_error<ignore_error>&) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T))
+BOOST_MATH_GPU_ENABLED BOOST_MATH_FORCEINLINE constexpr bool check_overflow(std::complex<T> /* val */, R* /* result */, const char* /* function */, const overflow_error<ignore_error>&) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T))
 { return false; }
 template <class R, class T>
-BOOST_MATH_FORCEINLINE constexpr bool check_underflow(T /* val */, R* /* result */, const char* /* function */, const underflow_error<ignore_error>&) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T))
+BOOST_MATH_GPU_ENABLED BOOST_MATH_FORCEINLINE constexpr bool check_underflow(T /* val */, R* /* result */, const char* /* function */, const underflow_error<ignore_error>&) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T))
 { return false; }
 template <class R, class T>
-BOOST_MATH_FORCEINLINE constexpr bool check_underflow(std::complex<T> /* val */, R* /* result */, const char* /* function */, const underflow_error<ignore_error>&) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T))
+BOOST_MATH_GPU_ENABLED BOOST_MATH_FORCEINLINE constexpr bool check_underflow(std::complex<T> /* val */, R* /* result */, const char* /* function */, const underflow_error<ignore_error>&) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T))
 { return false; }
 template <class R, class T>
-BOOST_MATH_FORCEINLINE constexpr bool check_denorm(T /* val */, R* /* result*/, const char* /* function */, const denorm_error<ignore_error>&) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T))
+BOOST_MATH_GPU_ENABLED BOOST_MATH_FORCEINLINE constexpr bool check_denorm(T /* val */, R* /* result*/, const char* /* function */, const denorm_error<ignore_error>&) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T))
 { return false; }
 template <class R, class T>
-BOOST_MATH_FORCEINLINE constexpr bool check_denorm(std::complex<T> /* val */, R* /* result*/, const char* /* function */, const denorm_error<ignore_error>&) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T))
+BOOST_MATH_GPU_ENABLED BOOST_MATH_FORCEINLINE constexpr bool check_denorm(std::complex<T> /* val */, R* /* result*/, const char* /* function */, const denorm_error<ignore_error>&) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T))
 { return false; }
 
 } // namespace detail
 
 template <class R, class Policy, class T>
-BOOST_MATH_FORCEINLINE R checked_narrowing_cast(T val, const char* function) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T) && is_noexcept_error_policy<Policy>::value)
+BOOST_MATH_GPU_ENABLED BOOST_MATH_FORCEINLINE R checked_narrowing_cast(T val, const char* function) noexcept(BOOST_MATH_IS_FLOAT(R) && BOOST_MATH_IS_FLOAT(T) && is_noexcept_error_policy<Policy>::value)
 {
    typedef typename Policy::overflow_error_type overflow_type;
    typedef typename Policy::underflow_error_type underflow_type;
@@ -852,7 +871,7 @@ BOOST_MATH_FORCEINLINE R checked_narrowing_cast(T val, const char* function) noe
 }
 
 template <class T, class Policy>
-inline void check_series_iterations(const char* function, std::uintmax_t max_iter, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(T) && is_noexcept_error_policy<Policy>::value)
+BOOST_MATH_GPU_ENABLED inline void check_series_iterations(const char* function, std::uintmax_t max_iter, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(T) && is_noexcept_error_policy<Policy>::value)
 {
    if(max_iter >= policies::get_max_series_iterations<Policy>())
       raise_evaluation_error<T>(
@@ -861,7 +880,7 @@ inline void check_series_iterations(const char* function, std::uintmax_t max_ite
 }
 
 template <class T, class Policy>
-inline void check_root_iterations(const char* function, std::uintmax_t max_iter, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(T) && is_noexcept_error_policy<Policy>::value)
+BOOST_MATH_GPU_ENABLED inline void check_root_iterations(const char* function, std::uintmax_t max_iter, const Policy& pol) noexcept(BOOST_MATH_IS_FLOAT(T) && is_noexcept_error_policy<Policy>::value)
 {
    if(max_iter >= policies::get_max_root_iterations<Policy>())
       raise_evaluation_error<T>(
@@ -871,25 +890,169 @@ inline void check_root_iterations(const char* function, std::uintmax_t max_iter,
 
 } //namespace policies
 
-namespace detail{
-
-//
-// Simple helper function to assist in returning a pair from a single value,
-// that value usually comes from one of the error handlers above:
-//
-template <class T>
-std::pair<T, T> pair_from_single(const T& val) BOOST_MATH_NOEXCEPT(T)
-{
-   return std::make_pair(val, val);
-}
-
-}
-
 #ifdef _MSC_VER
 #  pragma warning(pop)
 #endif
 
 }} // namespaces boost/math
 
+#else // Special values for NVRTC
+
+namespace boost {
+namespace math {
+namespace policies {
+
+template <class T, class Policy>
+BOOST_MATH_GPU_ENABLED constexpr T raise_domain_error(
+           const char* ,
+           const char* ,
+           const T& ,
+           const Policy&) BOOST_MATH_NOEXCEPT(T)
+{
+   // This may or may not do the right thing, but the user asked for the error
+   // to be ignored so here we go anyway:
+   return boost::math::numeric_limits<T>::quiet_NaN();
+}
+
+template <class T, class Policy>
+BOOST_MATH_GPU_ENABLED constexpr T raise_pole_error(
+           const char* function,
+           const char* message,
+           const T& val,
+           const  Policy&) BOOST_MATH_NOEXCEPT(T)
+{
+   return boost::math::numeric_limits<T>::quiet_NaN();
+}
+
+template <class T, class Policy>
+BOOST_MATH_GPU_ENABLED constexpr T raise_overflow_error(
+           const char* ,
+           const char* ,
+           const Policy&) BOOST_MATH_NOEXCEPT(T)
+{
+   // This may or may not do the right thing, but the user asked for the error
+   // to be ignored so here we go anyway:
+   return boost::math::numeric_limits<T>::has_infinity ? boost::math::numeric_limits<T>::infinity() : (boost::math::numeric_limits<T>::max)();
+}
+
+template <class T, class Policy>
+BOOST_MATH_GPU_ENABLED constexpr T raise_overflow_error(
+           const char* ,
+           const char* ,
+           const T&,
+           const Policy&) BOOST_MATH_NOEXCEPT(T)
+{
+   // This may or may not do the right thing, but the user asked for the error
+   // to be ignored so here we go anyway:
+   return boost::math::numeric_limits<T>::has_infinity ? boost::math::numeric_limits<T>::infinity() : (boost::math::numeric_limits<T>::max)();
+}
+
+template <class T, class Policy>
+BOOST_MATH_GPU_ENABLED constexpr T raise_underflow_error(
+           const char* ,
+           const char* ,
+           const Policy&) BOOST_MATH_NOEXCEPT(T)
+{
+   // This may or may not do the right thing, but the user asked for the error
+   // to be ignored so here we go anyway:
+   return static_cast<T>(0);
+}
+
+template <class T, class Policy>
+BOOST_MATH_GPU_ENABLED inline constexpr T raise_denorm_error(
+           const char* ,
+           const char* ,
+           const T& val,
+           const Policy&) BOOST_MATH_NOEXCEPT(T)
+{
+   // This may or may not do the right thing, but the user asked for the error
+   // to be ignored so here we go anyway:
+   return val;
+}
+
+template <class T, class Policy>
+BOOST_MATH_GPU_ENABLED constexpr T raise_evaluation_error(
+           const char* ,
+           const char* ,
+           const T& val,
+           const Policy&) BOOST_MATH_NOEXCEPT(T)
+{
+   // This may or may not do the right thing, but the user asked for the error
+   // to be ignored so here we go anyway:
+   return val;
+}
+
+template <class T, class TargetType, class Policy>
+BOOST_MATH_GPU_ENABLED constexpr TargetType raise_rounding_error(
+           const char* ,
+           const char* ,
+           const T& val,
+           const TargetType&,
+           const Policy&) BOOST_MATH_NOEXCEPT(T)
+{
+   // This may or may not do the right thing, but the user asked for the error
+   // to be ignored so here we go anyway:
+   static_assert(boost::math::numeric_limits<TargetType>::is_specialized, "The target type must have std::numeric_limits specialized.");
+   return  val > 0 ? (boost::math::numeric_limits<TargetType>::max)() : (boost::math::numeric_limits<TargetType>::is_integer ? (boost::math::numeric_limits<TargetType>::min)() : -(boost::math::numeric_limits<TargetType>::max)());
+}
+
+template <class T, class R, class Policy>
+BOOST_MATH_GPU_ENABLED inline constexpr T raise_indeterminate_result_error(
+           const char* ,
+           const char* ,
+           const T& ,
+           const R& result,
+           const Policy&) BOOST_MATH_NOEXCEPT(T)
+{
+   // This may or may not do the right thing, but the user asked for the error
+   // to be ignored so here we go anyway:
+   return result;
+}
+
+template <class R, class Policy, class T>
+BOOST_MATH_GPU_ENABLED BOOST_MATH_FORCEINLINE R checked_narrowing_cast(T val, const char* function) noexcept(boost::math::is_floating_point_v<R> && boost::math::is_floating_point_v<T>)
+{
+   // We only have ignore error policy so no reason to check
+   return static_cast<R>(val);
+}
+
+template <class T, class Policy>
+BOOST_MATH_GPU_ENABLED inline void check_series_iterations(const char* function, boost::math::uintmax_t max_iter, const Policy& pol) noexcept(boost::math::is_floating_point_v<T>)
+{
+   if(max_iter >= policies::get_max_series_iterations<Policy>())
+      raise_evaluation_error<T>(
+         function,
+         "Series evaluation exceeded %1% iterations, giving up now.", static_cast<T>(static_cast<double>(max_iter)), pol);
+}
+
+template <class T, class Policy>
+BOOST_MATH_GPU_ENABLED inline void check_root_iterations(const char* function, boost::math::uintmax_t max_iter, const Policy& pol) noexcept(boost::math::is_floating_point_v<T>)
+{
+   if(max_iter >= policies::get_max_root_iterations<Policy>())
+      raise_evaluation_error<T>(
+         function,
+         "Root finding evaluation exceeded %1% iterations, giving up now.", static_cast<T>(static_cast<double>(max_iter)), pol);
+}
+
+} // namespace policies
+} // namespace math
+} // namespace boost
+
+#endif // BOOST_MATH_HAS_NVRTC
+
+namespace boost { namespace math { namespace detail {
+
+//
+// Simple helper function to assist in returning a pair from a single value,
+// that value usually comes from one of the error handlers above:
+//
+template <class T>
+BOOST_MATH_GPU_ENABLED boost::math::pair<T, T> pair_from_single(const T& val) BOOST_MATH_NOEXCEPT(T)
+{
+   return boost::math::make_pair(val, val);
+}
+
+}}} // boost::math::detail
+
 #endif // BOOST_MATH_POLICY_ERROR_HANDLING_HPP
 
